@@ -29,19 +29,18 @@
 static MemoryContext cluster_mem;
 
 /*
- * Singly linked list of clusters.
+ * Tree of clusters.
  *
- * For searching by name.  If there will be lots of clusters
- * should use some faster search method, HTAB probably.
+ * For searching by name.
  */
-static ProxyCluster *cluster_list = NULL;
+static struct AATree cluster_tree;
 
 /*
  * Similar list for fake clusters (for CONNECT functions).
  *
  * Cluster name will be actual connect string.
  */
-static ProxyCluster *fake_cluster_list = NULL;
+static struct AATree fake_cluster_tree;
 
 /* plan for fetching cluster version */
 static void *version_plan;
@@ -89,6 +88,70 @@ check_valid_partcount(int n)
 	return (n > 0) && !(n & (n - 1));
 }
 
+static int cluster_name_cmp(uintptr_t val, struct AANode *node)
+{
+	const char *name = (const char *)val;
+	const ProxyCluster *cluster = container_of(node, ProxyCluster, node);
+
+	return strcmp(name, cluster->name);
+}
+
+static int conn_cstr_cmp(uintptr_t val, struct AANode *node)
+{
+	const char *name = (const char *)val;
+	const ProxyConnection *conn = container_of(node, ProxyConnection, node);
+
+	return strcmp(name, conn->connstr);
+}
+
+static void conn_free(struct AANode *node, void *arg)
+{
+	ProxyConnection *conn = container_of(node, ProxyConnection, node);
+
+	aatree_destroy(&conn->userstate_tree);
+	if (conn->res)
+		PQclear(conn->res);
+	pfree(conn);
+}
+
+static int state_user_cmp(uintptr_t val, struct AANode *node)
+{
+	const char *name = (const char *)val;
+	const ProxyConnectionState *state = container_of(node, ProxyConnectionState, node);
+
+	return strcmp(name, state->userinfo->username);
+}
+
+static void state_free(struct AANode *node, void *arg)
+{
+	ProxyConnectionState *state = container_of(node, ProxyConnectionState, node);
+
+	plproxy_disconnect(state);
+	memset(state, 0, sizeof(*state));
+	pfree(state);
+}
+
+static int userinfo_cmp(uintptr_t val, struct AANode *node)
+{
+	const char *name = (const char *)val;
+	const ConnUserInfo *info = container_of(node, ConnUserInfo, node);
+
+	return strcmp(name, info->username);
+}
+
+static void userinfo_free(struct AANode *node, void *arg)
+{
+	ConnUserInfo *info = container_of(node, ConnUserInfo, node);
+	pfree(info->username);
+	if (info->extra_connstr)
+	{
+		memset(info->extra_connstr, 0, strlen(info->extra_connstr));
+		pfree(info->extra_connstr);
+	}
+	memset(info, 0, sizeof(*info));
+	pfree(info);
+}
+
 /*
  * Create cache memory area and prepare plans
  */
@@ -104,6 +167,8 @@ plproxy_cluster_cache_init(void)
 										ALLOCSET_SMALL_MINSIZE,
 										ALLOCSET_SMALL_INITSIZE,
 										ALLOCSET_SMALL_MAXSIZE);
+	aatree_init(&cluster_tree, cluster_name_cmp, NULL);
+	aatree_init(&fake_cluster_tree, cluster_name_cmp, NULL);
 }
 
 /* initialize plans on demand */
@@ -149,73 +214,48 @@ plproxy_cluster_plan_init(void)
 /*
  * Drop partition and connection data from cluster.
  */
+
 static void
 free_connlist(ProxyCluster *cluster)
 {
-	int			i;
-	ProxyConnection *conn;
+	aatree_destroy(&cluster->conn_tree);
 
-	for (i = 0; i < cluster->conn_count; i++)
-	{
-		conn = &cluster->conn_list[i];
-		if (conn->db)
-			PQfinish(conn->db);
-		if (conn->res)
-			PQclear(conn->res);
-		if (conn->connstr)
-			pfree((void *) conn->connstr);
-	}
 	pfree(cluster->part_map);
-	pfree(cluster->conn_list);
+	pfree(cluster->active_list);
 
 	cluster->part_map = NULL;
 	cluster->part_count = 0;
 	cluster->part_mask = 0;
-	cluster->conn_list = NULL;
-	cluster->conn_count = 0;
+	cluster->active_count = 0;
 }
 
 /*
  * Add new database connection if it does not exists.
  */
-static ProxyConnection *
-add_connection(ProxyCluster *cluster, char *connstr, int part_num)
+static void
+add_connection(ProxyCluster *cluster, const char *connstr, int part_num)
 {
-	int			i;
+	struct AANode *node;
 	ProxyConnection *conn = NULL;
-	char	   *username;
-	StringInfo	final;
-
-	final = makeStringInfo();
-	appendStringInfoString(final, connstr);
-
-	/* append current user if not specified in connstr */
-	if (strstr(connstr, "user=") == NULL)
-	{
-		username = GetUserNameFromId(GetSessionUserId());
-		appendStringInfo(final, " user=%s", username);
-	}
 
 	/* check if already have it */
-	for (i = 0; i < cluster->conn_count && !conn; i++)
-	{
-		ProxyConnection *c = &cluster->conn_list[i];
-
-		if (strcmp(c->connstr, final->data) == 0)
-			conn = c;
-	}
+	node = aatree_search(&cluster->conn_tree, (uintptr_t)connstr);
+	if (node)
+		conn = container_of(node, ProxyConnection, node);
 
 	/* add new connection */
 	if (!conn)
 	{
-		conn = &cluster->conn_list[cluster->conn_count++];
-		conn->connstr = MemoryContextStrdup(cluster_mem, final->data);
+		conn = MemoryContextAllocZero(cluster_mem, sizeof(ProxyConnection));
+		conn->connstr = MemoryContextStrdup(cluster_mem, connstr);
 		conn->cluster = cluster;
+
+		aatree_init(&conn->userstate_tree, state_user_cmp, state_free);
+
+		aatree_insert(&cluster->conn_tree, (uintptr_t)connstr, &conn->node);
 	}
 
 	cluster->part_map[part_num] = conn;
-
-	return conn;
 }
 
 /*
@@ -248,6 +288,13 @@ get_version(ProxyFunction *func, Datum dname)
 	return DatumGetInt32(bin_val);
 }
 
+/* forget old values */
+static void
+clear_config(ProxyConfig *cf)
+{
+	memset(cf, 0, sizeof(*cf));
+}
+
 /* set a configuration option. */
 static void
 set_config_key(ProxyFunction *func, ProxyConfig *cf, const char *key, const char *val)
@@ -266,6 +313,8 @@ set_config_key(ProxyFunction *func, ProxyConfig *cf, const char *key, const char
 		cf->keepintvl = atoi(val);
 	else if (pg_strcasecmp("keepalive_count", key) == 0)
 		cf->keepcnt = atoi(val);
+	else if (pg_strcasecmp("default_user", key) == 0)
+		snprintf(cf->default_user, sizeof(cf->default_user), "%s", val);
 	else
 		plproxy_error(func, "Unknown config param: %s", key);
 }
@@ -296,6 +345,8 @@ get_config(ProxyCluster *cluster, Datum dname, ProxyFunction *func)
 	if (SPI_gettypeid(desc, 2) != TEXTOID)
 		plproxy_error(func, "Config column 2 must be text");
 
+	clear_config(&cluster->config);
+
 	/* fill values */
 	for (i = 0; i < SPI_processed; i++)
 	{
@@ -322,7 +373,7 @@ allocate_cluster_partitions(ProxyCluster *cluster, int nparts)
 	MemoryContext old_ctx;
 
 	/* free old one */
-	if (cluster->conn_list)
+	if (cluster->part_map)
 		free_connlist(cluster);
 
 	cluster->part_count = nparts;
@@ -330,9 +381,8 @@ allocate_cluster_partitions(ProxyCluster *cluster, int nparts)
 
 	/* allocate lists */
 	old_ctx = MemoryContextSwitchTo(cluster_mem);
-
 	cluster->part_map = palloc0(nparts * sizeof(ProxyConnection *));
-	cluster->conn_list = palloc0(nparts * sizeof(ProxyConnection));
+	cluster->active_list = palloc0(nparts * sizeof(ProxyConnection *));
 	MemoryContextSwitchTo(old_ctx);
 }
 
@@ -499,6 +549,78 @@ plproxy_fdw_validator(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(true);
 }
 
+static void
+reload_sqlmed_user(ProxyFunction *func, ProxyCluster *cluster)
+{
+	ConnUserInfo *userinfo = cluster->cur_userinfo;
+
+	UserMapping        *um;
+	HeapTuple			tup;
+	StringInfoData      cstr;
+	ListCell		   *cell;
+	AclResult			aclresult;
+	bool				got_user;
+
+
+	um = GetUserMapping(userinfo->user_oid, cluster->sqlmed_server_oid);
+
+	/* retry same lookup so we can set cache stamp... */
+    tup = SearchSysCache(USERMAPPINGUSERSERVER,
+						 ObjectIdGetDatum(um->userid),
+						 ObjectIdGetDatum(um->serverid),
+						 0, 0);
+    if (!HeapTupleIsValid(tup))
+	{
+		/* Specific mapping not found, try PUBLIC */
+		tup = SearchSysCache(USERMAPPINGUSERSERVER,
+							 ObjectIdGetDatum(InvalidOid),
+							 ObjectIdGetDatum(um->serverid),
+							 0, 0);
+		if (!HeapTupleIsValid(tup))
+			elog(ERROR, "cache lookup failed for user mapping (%u,%u)",
+				um->userid, um->serverid);
+	}
+	scstamp_set(USERMAPPINGOID, &userinfo->umStamp, tup);
+	ReleaseSysCache(tup);
+
+	/*
+	 * Check permissions, user must have usage on the server.
+	 */
+	aclresult = pg_foreign_server_aclcheck(um->serverid, um->userid, ACL_USAGE);
+	if (aclresult != ACLCHECK_OK)
+		aclcheck_error(aclresult, ACL_KIND_FOREIGN_SERVER, cluster->name);
+
+	/* Extract the common connect string elements from user mapping */
+	got_user = false;
+	initStringInfo(&cstr);
+	foreach(cell, um->options)
+	{
+		DefElem    *def = lfirst(cell);
+
+		if (strcmp(def->defname, "user") == 0)
+			got_user = true;
+
+		appendStringInfo(&cstr, " %s='%s'", def->defname, strVal(def->arg));
+	}
+
+	/* make sure we have 'user=' in connect string */
+	if (!got_user)
+		appendStringInfo(&cstr, " user='%s'", userinfo->username);
+
+	/* free old string */
+	if (userinfo->extra_connstr)
+	{
+		memset(userinfo->extra_connstr, 0, strlen(userinfo->extra_connstr));
+		pfree(userinfo->extra_connstr);
+		userinfo->extra_connstr = NULL;
+	}
+
+	/* set up new connect string */
+	userinfo->extra_connstr = MemoryContextStrdup(cluster_mem, cstr.data);
+	memset(cstr.data, 0, cstr.len);
+	pfree(cstr.data);
+}
+
 /*
  * Reload the cluster configuration and partitions from SQL/MED catalogs.
  */
@@ -506,19 +628,15 @@ static void
 reload_sqlmed_cluster(ProxyFunction *func, ProxyCluster *cluster,
 					  ForeignServer *foreign_server)
 {
-	UserMapping        *user_mapping;
+	ConnUserInfo		*info = cluster->cur_userinfo;
 	ForeignDataWrapper *fdw;
 	HeapTuple			tup;
 	AclResult			aclresult;
-	StringInfo          user_options;
 	ListCell		   *cell;
-	Oid					userid;
 	int					part_count = 0;
 	int					part_num;
 
 
-	userid = GetSessionUserId();
-	user_mapping = GetUserMapping(userid, foreign_server->serverid);
 	fdw = GetForeignDataWrapper(foreign_server->fdwid);
 
 	/*
@@ -531,44 +649,18 @@ reload_sqlmed_cluster(ProxyFunction *func, ProxyCluster *cluster,
     if (!HeapTupleIsValid(tup))
         elog(ERROR, "cache lookup failed for foreign server %u", foreign_server->serverid);
 
-	cluster->clusterTupleId = tup->t_self;
-	ReleaseSysCache(tup);
-
-    tup = SearchSysCache(USERMAPPINGUSERSERVER,
-						 ObjectIdGetDatum(user_mapping->userid),
-						 ObjectIdGetDatum(foreign_server->serverid),
-						 0, 0);
-
-    if (!HeapTupleIsValid(tup))
-	{
-		/* Specific mapping not found, try PUBLIC */
-		tup = SearchSysCache(USERMAPPINGUSERSERVER,
-							 ObjectIdGetDatum(InvalidOid),
-							 ObjectIdGetDatum(foreign_server->serverid),
-							 0, 0);
-		if (!HeapTupleIsValid(tup))
-			elog(ERROR, "cache lookup failed for user mapping (%u,%u)",
-				user_mapping->userid, foreign_server->serverid);
-	}
-
-	cluster->umTupleId = tup->t_self;
+	scstamp_set(FOREIGNSERVEROID, &cluster->clusterStamp, tup);
 	ReleaseSysCache(tup);
 
 	/*
 	 * Check permissions, user must have usage on the server.
 	 */
-	aclresult = pg_foreign_server_aclcheck(foreign_server->serverid, userid, ACL_USAGE);
+	aclresult = pg_foreign_server_aclcheck(foreign_server->serverid, info->user_oid, ACL_USAGE);
 	if (aclresult != ACLCHECK_OK)
 		aclcheck_error(aclresult, ACL_KIND_FOREIGN_SERVER, foreign_server->servername);
 
-	/* Extract the common connect string elements from user mapping */
-	user_options = makeStringInfo();
-	foreach(cell, user_mapping->options)
-	{
-		DefElem    *def = lfirst(cell);
-
-		appendStringInfo(user_options, "%s='%s' ", def->defname, strVal(def->arg));
-	}
+	/* drop old config values */
+	clear_config(&cluster->config);
 
 	/*
 	 * Collect the configuration definitions from foreign data wrapper.
@@ -612,15 +704,11 @@ reload_sqlmed_cluster(ProxyFunction *func, ProxyCluster *cluster,
 	foreach(cell, foreign_server->options)
 	{
 		DefElem    *def = lfirst(cell);
-		StringInfo  buf = makeStringInfo();
 
 		if (!extract_part_num(def->defname, &part_num))
 			continue;
 
-		appendStringInfo(buf, "%s%s%s", strVal(def->arg),
-				user_options->len ? " " : "",
-				user_options->data);
-		add_connection(cluster, buf->data, part_num);
+		add_connection(cluster, strVal(def->arg), part_num);
 	}
 }
 
@@ -676,6 +764,54 @@ determine_compat_mode(ProxyCluster *cluster)
 		elog(ERROR, "Pl/Proxy: cluster not found: %s", cluster->name);
 }
 
+static void inval_one_umap(struct AANode *n, void *arg)
+{
+	ConnUserInfo *info = container_of(n, ConnUserInfo, node);
+	SCInvalArg newStamp;
+
+	if (info->needs_reload)
+		/* already invalidated */
+		return;
+
+	if (arg == NULL)
+	{
+		info->needs_reload = true;
+		return;
+	}
+
+	newStamp = *(SCInvalArg *)arg;
+	if (scstamp_check(USERMAPPINGOID, &info->umStamp, newStamp))
+		/* user mappings changed */
+		info->needs_reload = true;
+}
+
+static void inval_umapping(struct AANode *n, void *arg)
+{
+	ProxyCluster *cluster = container_of(n, ProxyCluster, node);
+
+	aatree_walk(&cluster->userinfo_tree, AA_WALK_IN_ORDER, inval_one_umap, arg);
+}
+
+static void inval_fserver(struct AANode *n, void *arg)
+{
+	ProxyCluster *cluster = container_of(n, ProxyCluster, node);
+	SCInvalArg newStamp = *(SCInvalArg *)arg;
+
+	if (cluster->needs_reload)
+		/* already invalidated */
+		return;
+	else if (!cluster->sqlmed_cluster)
+		/* allow new SQL/MED servers to override compat definitions */
+		cluster->needs_reload = true;
+	else if (scstamp_check(FOREIGNSERVEROID, &cluster->clusterStamp, newStamp))
+		/* server definitions changed */
+		cluster->needs_reload = true;
+
+	/* tag all users too */
+	if (cluster->needs_reload)
+		inval_umapping(&cluster->node, NULL);
+}
+
 /*
  * Syscache inval callback function for foreign servers and user mappings.
  *
@@ -684,54 +820,32 @@ determine_compat_mode(ProxyCluster *cluster)
  * functions.
  */
 static void
-ClusterSyscacheCallback(Datum arg, int cacheid, ItemPointer tuplePtr)
+ClusterSyscacheCallback(Datum arg, int cacheid, SCInvalArg newStamp)
 {
-	ProxyCluster *cluster;
-
-	for (cluster = cluster_list; cluster; cluster = cluster->next)
-	{
-		if (cluster->needs_reload)
-		{
-			/* already invalidated */
-			continue;
-		}
-		else if (!tuplePtr)
-		{
-			/* invalidate all */
-			cluster->needs_reload = true;
-		}
-		else if (!cluster->sqlmed_cluster)
-		{
-			/* allow new SQL/MED servers to override compat definitions */
-			cluster->needs_reload = (cacheid == FOREIGNSERVEROID);
-		}
-		else if (cacheid == USERMAPPINGOID)
-		{
-			/* user mappings changed */
-			cluster->needs_reload = ItemPointerEquals(tuplePtr, &cluster->umTupleId);
-		}
-		else if (cacheid == FOREIGNSERVEROID)
-		{
-			/* server definitions changed */
-			cluster->needs_reload = ItemPointerEquals(tuplePtr, &cluster->clusterTupleId);
-		}
-	}
+	if (cacheid == FOREIGNSERVEROID)
+		aatree_walk(&cluster_tree, AA_WALK_IN_ORDER, inval_fserver, &newStamp);
+	else if (cacheid == USERMAPPINGOID)
+		aatree_walk(&cluster_tree, AA_WALK_IN_ORDER, inval_umapping, &newStamp);
 }
 
-#endif
-
 /*
  * Register syscache invalidation callbacks for SQL/MED clusters.
  */
 void
 plproxy_syscache_callback_init(void)
 {
-#ifdef PLPROXY_USE_SQLMED
 	CacheRegisterSyscacheCallback(FOREIGNSERVEROID, ClusterSyscacheCallback, (Datum) 0);
 	CacheRegisterSyscacheCallback(USERMAPPINGOID, ClusterSyscacheCallback, (Datum) 0);
-#endif
 }
 
+#else /* !PLPROXY_USE_SQLMED */
+
+void plproxy_syscache_callback_init(void) {}
+
+#endif
+
+
+
 /*
  * Reload the cluster configuration and partitions from plproxy.get_cluster*
  * functions.
@@ -767,19 +881,139 @@ new_cluster(const char *name)
 
 	cluster = palloc0(sizeof(*cluster));
 	cluster->name = pstrdup(name);
-	cluster->needs_reload = true;
+
+	aatree_init(&cluster->conn_tree, conn_cstr_cmp, conn_free);
+	aatree_init(&cluster->userinfo_tree, userinfo_cmp, userinfo_free);
 
 	MemoryContextSwitchTo(old_ctx);
 
 	return cluster;
 }
 
+/*
+ * Invalidate all connections for particular user
+ */
+#ifdef PLPROXY_USE_SQLMED
+
+static void inval_userinfo_state(struct AANode *node, void *arg)
+{
+	ProxyConnectionState *cur = container_of(node, ProxyConnectionState, node);
+	ConnUserInfo *userinfo = arg;
+
+	if (cur->userinfo == userinfo && cur->db)
+		plproxy_disconnect(cur);
+}
+
+static void inval_userinfo_conn(struct AANode *node, void *arg)
+{
+	ProxyConnection *conn = container_of(node, ProxyConnection, node);
+	ConnUserInfo *userinfo = arg;
+
+	aatree_walk(&conn->userstate_tree, AA_WALK_IN_ORDER, inval_userinfo_state, userinfo);
+}
+
+static void inval_user_connections(ProxyCluster *cluster, ConnUserInfo *userinfo)
+{
+	/* find all connections with this user and drop them */
+	aatree_walk(&cluster->conn_tree, AA_WALK_IN_ORDER, inval_userinfo_conn, userinfo);
+
+	/*
+	 * We can clear the flag only when it's certain
+	 * that no connections with old info exist
+	 */
+	userinfo->needs_reload = false;
+}
+
+#endif
+
+/*
+ * Initialize user info struct
+ */
+
+static ConnUserInfo *
+get_userinfo(ProxyCluster *cluster, Oid user_oid)
+{
+	ConnUserInfo *userinfo;
+	struct AANode *node;
+	const char *username;
+
+	username = GetUserNameFromId(user_oid);
+
+	node = aatree_search(&cluster->userinfo_tree, (uintptr_t)username);
+	if (node) {
+		userinfo = container_of(node, ConnUserInfo, node);
+	} else {
+		userinfo = MemoryContextAllocZero(cluster_mem, sizeof(*userinfo));
+		userinfo->username = MemoryContextStrdup(cluster_mem, username);
+
+		aatree_insert(&cluster->userinfo_tree, (uintptr_t)username, &userinfo->node);
+	}
+
+	if (userinfo->user_oid != user_oid)
+	{
+		/* user got renamed? */
+		userinfo->user_oid = user_oid;
+		userinfo->needs_reload = true;
+	}
+
+	return userinfo;
+}
+
+#if PG_VERSION_NUM < 90100
+
+static Oid
+get_role_oid(const char *rolname, bool missing_ok)
+{
+	Oid         oid;
+
+	oid = GetSysCacheOid(AUTHNAME, CStringGetDatum(rolname), 0, 0, 0);
+	if (!OidIsValid(oid) && !missing_ok)
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("role \"%s\" does not exist", rolname)));
+	return oid;
+}
+
+#endif
+
 /*
  * Refresh the cluster.
  */
 static void
 refresh_cluster(ProxyFunction *func, ProxyCluster *cluster)
 {
+	ConnUserInfo *uinfo;
+	ProxyConfig *cf = &cluster->config;
+	Oid user_oid = InvalidOid;
+
+	/*
+	 * Decide which user to use for connections.
+	 */
+	if (cf->default_user[0])
+	{
+		if (strcmp(cf->default_user, "session_user") == 0)
+			user_oid = GetSessionUserId();
+		else if (strcmp(cf->default_user, "current_user") == 0)
+			user_oid = GetUserId();
+		else if (1)
+			/* dont support custom users, seems unnecessary */
+			elog(ERROR, "default_user: Expect 'current_user' or 'session_user', got '%s'",
+				 cf->default_user);
+		else
+			/* easy to support, but seems confusing conceptually */
+			user_oid = get_role_oid(cf->default_user, false);
+	}
+	else
+	{
+		/* default: current_user */
+		user_oid = GetUserId();
+	}
+
+	/* set up user cache */
+	uinfo = get_userinfo(cluster, user_oid);
+	cluster->cur_userinfo = uinfo;
+
+	/* SQL/MED server reload */
 #ifdef PLPROXY_USE_SQLMED
 	if (cluster->needs_reload)
 	{
@@ -796,12 +1030,30 @@ refresh_cluster(ProxyFunction *func, ProxyCluster *cluster)
 		if (!cluster->sqlmed_cluster)
 			determine_compat_mode(cluster);
 		else
+		{
+			cluster->sqlmed_server_oid = server->serverid;
 			reload_sqlmed_cluster(func, cluster, server);
+		}
 	}
+
 #endif
 
-	/* Either no SQL/MED support or no such foreign server */
-	if (!cluster->sqlmed_cluster)
+	/* SQL/MED user reload */
+	if (uinfo->needs_reload)
+	{
+#ifdef PLPROXY_USE_SQLMED
+		if (cluster->sqlmed_cluster)
+		{
+			inval_user_connections(cluster, uinfo);
+			reload_sqlmed_user(func, cluster);
+		}
+		else
+#endif
+			uinfo->needs_reload = false;
+	}
+
+	/* old-style cluster reload */
+	if (!cluster->sqlmed_cluster && !cluster->fake_cluster)
 		reload_plproxy_cluster(func, cluster);
 
 	cluster->needs_reload = false;
@@ -814,43 +1066,37 @@ static ProxyCluster *
 fake_cluster(ProxyFunction *func, const char *connect_str)
 {
 	ProxyCluster *cluster;
-	ProxyConnection *conn;
 	MemoryContext old_ctx;
+	struct AANode *n;
 
 	/* search if cached */
-	for (cluster = fake_cluster_list; cluster; cluster = cluster->next)
+	n = aatree_search(&fake_cluster_tree, (uintptr_t)connect_str);
+	if (n)
 	{
-		if (strcmp(cluster->name, connect_str) == 0)
-			break;
+		cluster = container_of(n, ProxyCluster, node);
+		goto done;
 	}
 
-	if (cluster)
-		return cluster;
-
 	/* create if not */
+	cluster = new_cluster(connect_str);
 
 	old_ctx = MemoryContextSwitchTo(cluster_mem);
 
-	cluster = palloc0(sizeof(*cluster));
-	cluster->name = pstrdup(connect_str);
+	cluster->fake_cluster = true;
 	cluster->version = 1;
 	cluster->part_count = 1;
 	cluster->part_mask = 0;
-	cluster->conn_count = 1;
-	cluster->part_map = palloc(sizeof(ProxyConnection *));
-	cluster->conn_list = palloc0(sizeof(ProxyConnection));
-	conn = &cluster->conn_list[0];
-	conn->cluster = cluster;
-	cluster->part_map[0] = conn;
-
-	conn->connstr = pstrdup(cluster->name);
-	conn->state = C_NONE;
+	cluster->part_map = palloc(cluster->part_count * sizeof(ProxyConnection *));
+	cluster->active_list = palloc(cluster->part_count * sizeof(ProxyConnection *));
 
 	MemoryContextSwitchTo(old_ctx);
 
-	cluster->next = fake_cluster_list;
-	fake_cluster_list = cluster;
+	add_connection(cluster, connect_str, 0);
 
+	aatree_insert(&fake_cluster_tree, (uintptr_t)connect_str, &cluster->node);
+
+done:
+	refresh_cluster(func, cluster);
 	return cluster;
 }
 
@@ -891,8 +1137,9 @@ resolve_query(ProxyFunction *func, FunctionCallInfo fcinfo, ProxyQuery *query)
 ProxyCluster *
 plproxy_find_cluster(ProxyFunction *func, FunctionCallInfo fcinfo)
 {
-	ProxyCluster *cluster;
+	ProxyCluster *cluster = NULL;
 	const char *name;
+	struct AANode *node;
 
 
 	/* functions used CONNECT with query */
@@ -914,18 +1161,16 @@ plproxy_find_cluster(ProxyFunction *func, FunctionCallInfo fcinfo)
 		name = func->cluster_name;
 
 	/* search if cached */
-	for (cluster = cluster_list; cluster; cluster = cluster->next)
-	{
-		if (strcmp(cluster->name, name) == 0)
-			break;
-	}
+	node = aatree_search(&cluster_tree, (uintptr_t)name);
+	if (node)
+		cluster = container_of(node, ProxyCluster, node);
 
 	/* create if not */
 	if (!cluster)
 	{
 		cluster = new_cluster(name);
-		cluster->next = cluster_list;
-		cluster_list = cluster;
+		cluster->needs_reload = true;
+		aatree_insert(&cluster_tree, (uintptr_t)name, &cluster->node);
 	}
 
 	/* determine cluster type, reload parts if necessary */
@@ -934,61 +1179,110 @@ plproxy_find_cluster(ProxyFunction *func, FunctionCallInfo fcinfo)
 	return cluster;
 }
 
-static void
-clean_cluster(ProxyCluster *cluster, struct timeval * now)
+/*
+ * Move connection to active list and init current
+ * connection state.
+ */
+void plproxy_activate_connection(struct ProxyConnection *conn)
 {
-	ProxyConnection *conn;
-	ProxyConfig *cf = &cluster->config;
-	time_t		age;
-	int			i;
-	bool		drop;
+	ProxyCluster *cluster = conn->cluster;
+	ConnUserInfo *userinfo = cluster->cur_userinfo;
+	const char *username = userinfo->username;
+	struct AANode *node;
+	ProxyConnectionState *cur;
 
-	for (i = 0; i < cluster->conn_count; i++)
-	{
-		conn = &cluster->conn_list[i];
-		if (conn->res)
-		{
-			PQclear(conn->res);
-			conn->res = NULL;
-		}
-		if (!conn->db)
-			continue;
+	/* move connection to active_list */
+	cluster->active_list[cluster->active_count] = conn;
+	cluster->active_count++;
 
-		drop = false;
-		if (PQstatus(conn->db) != CONNECTION_OK)
-		{
-			drop = true;
-		}
-		else if (cf->connection_lifetime <= 0)
-		{
-			/* no aging */
-		}
-		else
-		{
-			age = now->tv_sec - conn->connect_time;
-			if (age >= cf->connection_lifetime)
-				drop = true;
-		}
+	/* fill ->cur pointer */
 
-		if (drop)
-		{
-			PQfinish(conn->db);
-			conn->db = NULL;
-			conn->state = C_NONE;
-		}
+	node = aatree_search(&conn->userstate_tree, (uintptr_t)username);
+	if (node) {
+		cur = container_of(node, ProxyConnectionState, node);
+	} else {
+		cur = MemoryContextAllocZero(cluster_mem, sizeof(*cur));
+		cur->userinfo = userinfo;
+		aatree_insert(&conn->userstate_tree, (uintptr_t)username, &cur->node);
 	}
+	conn->cur = cur;
 }
 
 /*
  * Clean old connections and results from all clusters.
  */
+
+struct MaintInfo {
+	struct ProxyConfig *cf;
+	struct timeval *now;
+};
+
+static void clean_state(struct AANode *node, void *arg)
+{
+	ProxyConnectionState *cur = container_of(node, ProxyConnectionState, node);
+	ConnUserInfo *uinfo = cur->userinfo;
+	struct MaintInfo *maint = arg;
+	ProxyConfig *cf = maint->cf;
+	struct timeval *now = maint->now;
+	time_t		age;
+	bool		drop;
+
+	if (!cur->db)
+		return;
+
+	drop = false;
+	if (PQstatus(cur->db) != CONNECTION_OK)
+	{
+		drop = true;
+	}
+	else if (uinfo->needs_reload)
+	{
+		drop = true;
+	}
+	else if (cf->connection_lifetime <= 0)
+	{
+		/* no aging */
+	}
+	else
+	{
+		age = now->tv_sec - cur->connect_time;
+		if (age >= cf->connection_lifetime)
+			drop = true;
+	}
+
+	if (drop)
+		plproxy_disconnect(cur);
+}
+
+static void clean_conn(struct AANode *node, void *arg)
+{
+	ProxyConnection *conn = container_of(node, ProxyConnection, node);
+	struct MaintInfo *maint = arg;
+
+	if (conn->res)
+	{
+		PQclear(conn->res);
+		conn->res = NULL;
+	}
+
+	aatree_walk(&conn->userstate_tree, AA_WALK_IN_ORDER, clean_state, maint);
+}
+
+static void clean_cluster(struct AANode *n, void *arg)
+{
+	ProxyCluster *cluster = container_of(n, ProxyCluster, node);
+	struct MaintInfo maint;
+
+	maint.cf = &cluster->config;
+	maint.now = arg;
+
+	aatree_walk(&cluster->conn_tree, AA_WALK_IN_ORDER, clean_conn, &maint);
+}
+
 void
 plproxy_cluster_maint(struct timeval * now)
 {
-	ProxyCluster *cluster;
-
-	for (cluster = cluster_list; cluster; cluster = cluster->next)
-		clean_cluster(cluster, now);
-	for (cluster = fake_cluster_list; cluster; cluster = cluster->next)
-		clean_cluster(cluster, now);
+	aatree_walk(&cluster_tree, AA_WALK_IN_ORDER, clean_cluster, now);
+	aatree_walk(&fake_cluster_tree, AA_WALK_IN_ORDER, clean_cluster, now);
 }
+
