@@ -67,7 +67,7 @@ static void
 conn_error(ProxyFunction *func, ProxyConnection *conn, const char *desc)
 {
 	plproxy_error(func, "[%s] %s: %s",
-				  PQdb(conn->db), desc, PQerrorMessage(conn->db));
+				  PQdb(conn->cur->db), desc, PQerrorMessage(conn->cur->db));
 }
 
 /* Compare if major/minor match. Works on "MAJ.MIN.*" */
@@ -102,13 +102,13 @@ flush_connection(ProxyFunction *func, ProxyConnection *conn)
 	int res;
 
 	/* flush it down */
-	res = PQflush(conn->db);
+	res = PQflush(conn->cur->db);
 
 	/* set actual state */
 	if (res > 0)
-		conn->state = C_QUERY_WRITE;
+		conn->cur->state = C_QUERY_WRITE;
 	else if (res == 0)
-		conn->state = C_QUERY_READ;
+		conn->cur->state = C_QUERY_READ;
 	else
 		conn_error(func, conn, "PQflush");
 }
@@ -130,14 +130,14 @@ tune_connection(ProxyFunction *func, ProxyConnection *conn)
 	/*
 	 * check if target server has same backend version.
 	 */
-	dst_ver = PQparameterStatus(conn->db, "server_version");
-	conn->same_ver = cmp_branch(dst_ver, PG_VERSION);
+	dst_ver = PQparameterStatus(conn->cur->db, "server_version");
+	conn->cur->same_ver = cmp_branch(dst_ver, PG_VERSION);
 
 	/*
 	 * Make sure remote I/O is done using local server_encoding.
 	 */
 	this_enc = GetDatabaseEncodingName();
-	dst_enc = PQparameterStatus(conn->db, "client_encoding");
+	dst_enc = PQparameterStatus(conn->cur->db, "client_encoding");
 	if (dst_enc && strcmp(this_enc, dst_enc))
 	{
 		if (!sql)
@@ -148,7 +148,7 @@ tune_connection(ProxyFunction *func, ProxyConnection *conn)
 	/*
 	 * if second time in this function, they should be active already.
 	 */
-	if (sql && conn->tuning)
+	if (sql && conn->cur->tuning)
 	{
 		/* display SET query */
 		appendStringInfo(sql, "-- does not seem to apply");
@@ -160,9 +160,9 @@ tune_connection(ProxyFunction *func, ProxyConnection *conn)
 	 */
 	if (sql)
 	{
-		conn->tuning = 1;
-		conn->state = C_QUERY_WRITE;
-		if (!PQsendQuery(conn->db, sql->data))
+		conn->cur->tuning = 1;
+		conn->cur->state = C_QUERY_WRITE;
+		if (!PQsendQuery(conn->cur->db, sql->data))
 			conn_error(func, conn, "PQsendQuery");
 		pfree(sql->data);
 		pfree(sql);
@@ -171,7 +171,7 @@ tune_connection(ProxyFunction *func, ProxyConnection *conn)
 		return 1;
 	}
 
-	conn->tuning = 0;
+	conn->cur->tuning = 0;
 	return 0;
 }
 
@@ -187,14 +187,14 @@ send_query(ProxyFunction *func, ProxyConnection *conn,
 	int			binary_result = 0;
 
 	gettimeofday(&now, NULL);
-	conn->query_time = now.tv_sec;
+	conn->cur->query_time = now.tv_sec;
 
 	tune_connection(func, conn);
-	if (conn->tuning)
+	if (conn->cur->tuning)
 		return;
 
 	/* use binary result only on same backend ver */
-	if (cf->disable_binary == 0 && conn->same_ver)
+	if (cf->disable_binary == 0 && conn->cur->same_ver)
 	{
 		/* binary recv for non-record types */
 		if (func->ret_scalar)
@@ -210,8 +210,8 @@ send_query(ProxyFunction *func, ProxyConnection *conn,
 	}
 
 	/* send query */
-	conn->state = C_QUERY_WRITE;
-	res = PQsendQueryParams(conn->db, q->sql, q->arg_count,
+	conn->cur->state = C_QUERY_WRITE;
+	res = PQsendQueryParams(conn->cur->db, q->sql, q->arg_count,
 							NULL,		/* paramTypes */
 							values,		/* paramValues */
 							plengths,	/* paramLengths */
@@ -233,19 +233,19 @@ check_old_conn(ProxyFunction *func, ProxyConnection *conn, struct timeval * now)
 	struct pollfd	pfd;
 	ProxyConfig *cf = &func->cur_cluster->config;
 
-	if (PQstatus(conn->db) != CONNECTION_OK)
+	if (PQstatus(conn->cur->db) != CONNECTION_OK)
 		return false;
 
 	/* check if too old */
 	if (cf->connection_lifetime > 0)
 	{
-		t = now->tv_sec - conn->connect_time;
+		t = now->tv_sec - conn->cur->connect_time;
 		if (t >= cf->connection_lifetime)
 			return false;
 	}
 
 	/* how long ts been idle */
-	t = now->tv_sec - conn->query_time;
+	t = now->tv_sec - conn->cur->query_time;
 	if (t < PLPROXY_IDLE_CONN_CHECK)
 		return true;
 
@@ -254,7 +254,7 @@ check_old_conn(ProxyFunction *func, ProxyConnection *conn, struct timeval * now)
 	 * are events pending.  If there are drop the connection.
 	 */
 intr_loop:
-	pfd.fd = PQsocket(conn->db);
+	pfd.fd = PQsocket(conn->cur->db);
 	pfd.events = POLLIN;
 	pfd.revents = 0;
 
@@ -352,7 +352,7 @@ static void setup_keepalive(ProxyConnection *conn)
 {
 	struct sockaddr sa;
 	socklen_t salen = sizeof(sa);
-	int fd = PQsocket(conn->db);
+	int fd = PQsocket(conn->cur->db);
 	ProxyConfig *config = &conn->cluster->config;
 
 	/* turn on keepalive */
@@ -375,19 +375,37 @@ handle_notice(void *arg, const PGresult *res)
 	plproxy_remote_error(cluster->cur_func, conn, res, false);
 }
 
+static const char *
+get_connstr(ProxyConnection *conn)
+{
+	StringInfoData cstr;
+	ConnUserInfo *info = conn->cluster->cur_userinfo;
+
+	if (strstr(conn->connstr, "user=") != NULL)
+		return pstrdup(conn->connstr);
+
+	initStringInfo(&cstr);
+	if (info->extra_connstr)
+		appendStringInfo(&cstr, "%s %s", conn->connstr, info->extra_connstr);
+	else
+		appendStringInfo(&cstr, "%s user='%s'", conn->connstr, info->username);
+	return cstr.data;
+}
+
 /* check existing conn status or launch new conn */
 static void
 prepare_conn(ProxyFunction *func, ProxyConnection *conn)
 {
 	struct timeval now;
+	const char *connstr;
 
 	gettimeofday(&now, NULL);
 
 	/* state should be C_READY or C_NONE */
-	switch (conn->state)
+	switch (conn->cur->state)
 	{
 		case C_DONE:
-			conn->state = C_READY;
+			conn->cur->state = C_READY;
 		case C_READY:
 			if (check_old_conn(func, conn, &now))
 				return;
@@ -398,29 +416,27 @@ prepare_conn(ProxyFunction *func, ProxyConnection *conn)
 		case C_QUERY_WRITE:
 			/* close rotten connection */
 			elog(NOTICE, "PL/Proxy: dropping stale conn");
-			PQfinish(conn->db);
-			conn->db = NULL;
-			conn->state = C_NONE;
-			conn->tuning = 0;
+			plproxy_disconnect(conn->cur);
 		case C_NONE:
 			break;
 	}
 
-	conn->connect_time = now.tv_sec;
+	conn->cur->connect_time = now.tv_sec;
 
 	/* launch new connection */
-	conn->db = PQconnectStart(conn->connstr);
-	if (conn->db == NULL)
+	connstr = get_connstr(conn);
+	conn->cur->db = PQconnectStart(connstr);
+	if (conn->cur->db == NULL)
 		plproxy_error(func, "No memory for PGconn");
 
 	/* tag connection dirty */
-	conn->state = C_CONNECT_WRITE;
+	conn->cur->state = C_CONNECT_WRITE;
 
-	if (PQstatus(conn->db) == CONNECTION_BAD)
+	if (PQstatus(conn->cur->db) == CONNECTION_BAD)
 		conn_error(func, conn, "PQconnectStart");
 
 	/* override default notice handler */
-	PQsetNoticeReceiver(conn->db, handle_notice, conn);
+	PQsetNoticeReceiver(conn->cur->db, handle_notice, conn);
 
 	setup_keepalive(conn);
 }
@@ -437,13 +453,13 @@ another_result(ProxyFunction *func, ProxyConnection *conn)
 	PGresult   *res;
 
 	/* got one */
-	res = PQgetResult(conn->db);
+	res = PQgetResult(conn->cur->db);
 	if (res == NULL)
 	{
-		if (conn->tuning)
-			conn->state = C_READY;
+		if (conn->cur->tuning)
+			conn->cur->state = C_READY;
 		else
-			conn->state = C_DONE;
+			conn->cur->state = C_DONE;
 		return false;
 	}
 
@@ -489,21 +505,21 @@ handle_conn(ProxyFunction *func, ProxyConnection *conn)
 	int			res;
 	PostgresPollingStatusType poll_res;
 
-	switch (conn->state)
+	switch (conn->cur->state)
 	{
 		case C_CONNECT_READ:
 		case C_CONNECT_WRITE:
-			poll_res = PQconnectPoll(conn->db);
+			poll_res = PQconnectPoll(conn->cur->db);
 			switch (poll_res)
 			{
 				case PGRES_POLLING_WRITING:
-					conn->state = C_CONNECT_WRITE;
+					conn->cur->state = C_CONNECT_WRITE;
 					break;
 				case PGRES_POLLING_READING:
-					conn->state = C_CONNECT_READ;
+					conn->cur->state = C_CONNECT_READ;
 					break;
 				case PGRES_POLLING_OK:
-					conn->state = C_READY;
+					conn->cur->state = C_READY;
 					break;
 				case PGRES_POLLING_ACTIVE:
 				case PGRES_POLLING_FAILED:
@@ -514,7 +530,7 @@ handle_conn(ProxyFunction *func, ProxyConnection *conn)
 			flush_connection(func, conn);
 			break;
 		case C_QUERY_READ:
-			res = PQconsumeInput(conn->db);
+			res = PQconsumeInput(conn->cur->db);
 			if (res == 0)
 				conn_error(func, conn, "PQconsumeInput");
 
@@ -522,7 +538,7 @@ handle_conn(ProxyFunction *func, ProxyConnection *conn)
 			while (1)
 			{
 				/* if PQisBusy, then incomplete result */
-				if (PQisBusy(conn->db))
+				if (PQisBusy(conn->cur->db))
 					break;
 
 				/* got one */
@@ -556,10 +572,10 @@ poll_conns(ProxyFunction *func, ProxyCluster *cluster)
 	int numfds = 0;
 	int ev = 0;
 
-	if (pfd_allocated < cluster->conn_count)
+	if (pfd_allocated < cluster->active_count)
 	{
 		struct pollfd *tmp;
-		int num = cluster->conn_count;
+		int num = cluster->active_count;
 		if (num < 64)
 			num = 64;
 		if (pfd_cache == NULL)
@@ -572,14 +588,14 @@ poll_conns(ProxyFunction *func, ProxyCluster *cluster)
 		pfd_allocated = num;
 	}
 
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		conn = &cluster->conn_list[i];
+		conn = cluster->active_list[i];
 		if (!conn->run_tag)
 			continue;
 
 		/* decide what to do */
-		switch (conn->state)
+		switch (conn->cur->state)
 		{
 			case C_DONE:
 			case C_READY:
@@ -597,7 +613,7 @@ poll_conns(ProxyFunction *func, ProxyCluster *cluster)
 
 		/* add fd to proper set */
 		pf = pfd_cache + numfds++;
-		pf->fd = PQsocket(conn->db);
+		pf->fd = PQsocket(conn->cur->db);
 		pf->events = ev;
 		pf->revents = 0;
 	}
@@ -615,13 +631,13 @@ poll_conns(ProxyFunction *func, ProxyCluster *cluster)
 
 	/* now recheck the conns */
 	pf = pfd_cache;
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		conn = &cluster->conn_list[i];
+		conn = cluster->active_list[i];
 		if (!conn->run_tag)
 			continue;
 
-		switch (conn->state)
+		switch (conn->cur->state)
 		{
 			case C_DONE:
 			case C_READY:
@@ -637,7 +653,7 @@ poll_conns(ProxyFunction *func, ProxyCluster *cluster)
 		/*
 		 * they should be in same order as called,
 		 */
-		fd = PQsocket(conn->db);
+		fd = PQsocket(conn->cur->db);
 		if (pf->fd != fd)
 			elog(WARNING, "fd order from poll() is messed up?");
 
@@ -655,13 +671,13 @@ check_timeouts(ProxyFunction *func, ProxyCluster *cluster, ProxyConnection *conn
 {
 	ProxyConfig *cf = &cluster->config;
 
-	switch (conn->state)
+	switch (conn->cur->state)
 	{
 		case C_CONNECT_READ:
 		case C_CONNECT_WRITE:
 			if (cf->connect_timeout <= 0)
 				break;
-			if (now - conn->connect_time <= cf->connect_timeout)
+			if (now - conn->cur->connect_time <= cf->connect_timeout)
 				break;
 			plproxy_error(func, "connect timeout to: %s", conn->connstr);
 			break;
@@ -670,7 +686,7 @@ check_timeouts(ProxyFunction *func, ProxyCluster *cluster, ProxyConnection *conn
 		case C_QUERY_WRITE:
 			if (cf->query_timeout <= 0)
 				break;
-			if (now - conn->query_time <= cf->query_timeout)
+			if (now - conn->cur->query_time <= cf->query_timeout)
 				break;
 			plproxy_error(func, "query timeout");
 			break;
@@ -691,9 +707,9 @@ remote_execute(ProxyFunction *func)
 	struct timeval now;
 
 	/* either launch connection or send query */
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		conn = &cluster->conn_list[i];
+		conn = cluster->active_list[i];
 		if (!conn->run_tag)
 			continue;
 
@@ -702,7 +718,7 @@ remote_execute(ProxyFunction *func)
 		pending++;
 
 		/* if conn is ready, then send query away */
-		if (conn->state == C_READY)
+		if (conn->cur->state == C_READY)
 			send_query(func, conn, conn->param_values, conn->param_lengths, conn->param_formats);
 	}
 
@@ -719,17 +735,17 @@ remote_execute(ProxyFunction *func)
 		/* recheck */
 		pending = 0;
 		gettimeofday(&now, NULL);
-		for (i = 0; i < cluster->conn_count; i++)
+		for (i = 0; i < cluster->active_count; i++)
 		{
-			conn = &cluster->conn_list[i];
+			conn = cluster->active_list[i];
 			if (!conn->run_tag)
 				continue;
 
 			/* login finished, send query */
-			if (conn->state == C_READY)
+			if (conn->cur->state == C_READY)
 				send_query(func, conn, conn->param_values, conn->param_lengths, conn->param_formats);
 
-			if (conn->state != C_DONE)
+			if (conn->cur->state != C_DONE)
 				pending++;
 
 			check_timeouts(func, cluster, conn, now.tv_sec);
@@ -737,9 +753,9 @@ remote_execute(ProxyFunction *func)
 	}
 
 	/* review results, calculate total */
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		conn = &cluster->conn_list[i];
+		conn = cluster->active_list[i];
 
 		if ((conn->run_tag || conn->res)
 			&& !(conn->run_tag && conn->res))
@@ -748,7 +764,7 @@ remote_execute(ProxyFunction *func)
 		if (!conn->run_tag)
 			continue;
 
-		if (conn->state != C_DONE)
+		if (conn->cur->state != C_DONE)
 			plproxy_error(func, "Unfinished connection");
 		if (conn->res == NULL)
 			plproxy_error(func, "Lost result");
@@ -775,15 +791,15 @@ remote_cancel(ProxyFunction *func)
 	if (cluster == NULL)
 		return;
 
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		conn = &cluster->conn_list[i];
-		if (conn->state == C_NONE ||
-			conn->state == C_READY ||
-			conn->state == C_DONE)
+		conn = cluster->active_list[i];
+		if (conn->cur->state == C_NONE ||
+			conn->cur->state == C_READY ||
+			conn->cur->state == C_DONE)
 			continue;
 
-		cancel = PQgetCancel(conn->db);
+		cancel = PQgetCancel(conn->cur->db);
 		if (cancel == NULL)
 		{
 			elog(NOTICE, "Invalid connection!");
@@ -796,6 +812,20 @@ remote_cancel(ProxyFunction *func)
 	}
 }
 
+/*
+ * Tag & move tagged connections to active list
+ */
+
+static void tag_part(struct ProxyCluster *cluster, int i, int tag)
+{
+	ProxyConnection *conn = cluster->part_map[i];
+
+	if (!conn->run_tag)
+		plproxy_activate_connection(conn);
+
+	conn->run_tag = tag;
+}
+
 /*
  * Run hash function and tag connections. If any of the hash function 
  * arguments are mentioned in the split_arrays an element of the array
@@ -838,7 +868,7 @@ tag_hash_partitions(ProxyFunction *func, FunctionCallInfo fcinfo, int tag,
 			plproxy_error(func, "Hash result must be int2, int4 or int8");
 
 		hashval &= cluster->part_mask;
-		cluster->part_map[hashval]->run_tag = tag;
+		tag_part(cluster, hashval, tag);
 	}
 
 	/* sanity check */
@@ -889,17 +919,17 @@ tag_run_on_partitions(ProxyFunction *func, FunctionCallInfo fcinfo, int tag,
 			break;
 		case R_ALL:
 			for (i = 0; i < cluster->part_count; i++)
-				cluster->part_map[i]->run_tag = tag;
+				tag_part(cluster, i, tag);
 			break;
 		case R_EXACT:
 			i = func->exact_nr;
 			if (i < 0 || i >= cluster->part_count)
 				plproxy_error(func, "part number out of range");
-			cluster->part_map[i]->run_tag = tag;
+			tag_part(cluster, i, tag);
 			break;
 		case R_ANY:
 			i = random() & cluster->part_mask;
-			cluster->part_map[i]->run_tag = tag;
+			tag_part(cluster, i, tag);
 			break;
 		default:
 			plproxy_error(func, "uninitialized run_type");
@@ -979,9 +1009,9 @@ prepare_and_tag_partitions(ProxyFunction *func, FunctionCallInfo fcinfo)
 		tag_run_on_partitions(func, fcinfo, my_tag, arrays_to_split, row);
 
 		/* Add the array elements to the partitions tagged in previous step */
-		for (part = 0; part < cluster->conn_count; part++)
+		for (part = 0; part < cluster->active_count; part++)
 		{
-			ProxyConnection	   *conn = &cluster->conn_list[part];
+			ProxyConnection	   *conn = cluster->active_list[part];
 
 			if (conn->run_tag != my_tag)
 				continue;
@@ -1008,9 +1038,9 @@ prepare_and_tag_partitions(ProxyFunction *func, FunctionCallInfo fcinfo)
 	 * Finally, copy the accumulated arrays to the actual connections
 	 * to be used as parameters.
 	 */
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		ProxyConnection *conn = &cluster->conn_list[i];
+		ProxyConnection *conn = cluster->active_list[i];
 
 		if (!conn->run_tag)
 			continue;
@@ -1056,9 +1086,9 @@ prepare_query_parameters(ProxyFunction *func, FunctionCallInfo fcinfo)
 		}
 
 		/* Add the parameters to partitions */
-		for (part = 0; part < cluster->conn_count; part++)
+		for (part = 0; part < cluster->active_count; part++)
 		{
-			ProxyConnection *conn = &cluster->conn_list[part];
+			ProxyConnection *conn = cluster->active_list[part];
 
 			if (!conn->run_tag)
 				continue;
@@ -1103,9 +1133,9 @@ plproxy_clean_results(ProxyCluster *cluster)
 	cluster->ret_total = 0;
 	cluster->ret_cur_conn = 0;
 
-	for (i = 0; i < cluster->conn_count; i++)
+	for (i = 0; i < cluster->active_count; i++)
 	{
-		conn = &cluster->conn_list[i];
+		conn = cluster->active_list[i];
 		if (conn->res)
 		{
 			PQclear(conn->res);
@@ -1114,10 +1144,30 @@ plproxy_clean_results(ProxyCluster *cluster)
 		conn->pos = 0;
 		conn->run_tag = 0;
 		conn->bstate = NULL;
+		conn->cur = NULL;
+		cluster->active_list[i] = NULL;
 	}
+
+	/* reset active_list */
+	cluster->active_count = 0;
+
 	/* conn state checks are done in prepare_conn */
 }
 
+/* Drop one connection */
+void plproxy_disconnect(ProxyConnectionState *cur)
+{
+	if (cur->db)
+		PQfinish(cur->db);
+	cur->db = NULL;
+	cur->state = C_NONE;
+	cur->tuning = 0;
+	cur->connect_time = 0;
+	cur->query_time = 0;
+	cur->same_ver = 0;
+	cur->tuning = 0;
+}
+
 /* Select partitions and execute query on them */
 void
 plproxy_exec(ProxyFunction *func, FunctionCallInfo fcinfo)
