@@ -109,7 +109,7 @@ plproxy_remote_error(ProxyFunction *func, ProxyConnection *conn, const PGresult
 
 	ereport(elevel, (
 		errcode(MAKE_SQLSTATE(ss[0], ss[1], ss[2], ss[3], ss[4])),
-		errmsg("%s(%d): [%s] REMOTE %s: %s", func->name, func->arg_count, PQdb(conn->db), sev, msg),
+		errmsg("%s(%d): [%s] REMOTE %s: %s", func->name, func->arg_count, PQdb(conn->cur->db), sev, msg),
 		det ? errdetail("Remote detail: %s", det) : 0,
 		hint ? errhint("Remote hint: %s", hint) : 0,
 		spos ? errposition(atoi(spos)) : 0,
@@ -122,11 +122,11 @@ plproxy_remote_error(ProxyFunction *func, ProxyConnection *conn, const PGresult
  * Library load-time initialization.
  * Do the initialization when SPI is active to simplify the code.
  */
+static bool initialized = false;
+
 static void
 plproxy_startup_init(void)
 {
-	static bool initialized = false;
-
 	if (initialized)
 		return;
 
@@ -146,6 +146,9 @@ run_maint(void)
 	static struct timeval last = {0, 0};
 	struct timeval now;
 
+	if (!initialized)
+		return;
+
 	gettimeofday(&now, NULL);
 	if (now.tv_sec - last.tv_sec < 2 * 60)
 		return;
