@@ -54,6 +54,8 @@
 #include <utils/lsyscache.h>
 #include <utils/memutils.h>
 #include <utils/syscache.h>
+
+#include "aatree.h"
 #include "rowstamp.h"
 
 #include <libpq-fe.h>
@@ -62,6 +64,16 @@
 #error PL/Proxy requires 8.2+
 #endif
 
+/* give offset of a field inside struct */
+#ifndef offsetof
+#define offsetof(type, field) ((unsigned long)&(((type *)0)->field))
+#endif
+
+/* given pointer to field inside struct, return pointer to struct */
+#ifndef container_of
+#define container_of(ptr, type, field) ((type *)((char *)(ptr) - offsetof(type, field)))
+#endif
+
 /*
  * backwards compat with 8.2
  */
@@ -75,6 +87,14 @@
 #define PG_DETOAST_DATUM_PACKED(x) PG_DETOAST_DATUM(x)
 #endif
 
+#ifndef PG_PRINTF_ATTRIBUTE
+#ifdef WIN32
+#define PG_PRINTF_ATTRIBUTE gnu_printf
+#else
+#define PG_PRINTF_ATTRIBUTE printf
+#endif
+#endif
+
 /*
  * Determine if this argument is to SPLIT
  */
@@ -124,23 +144,47 @@ typedef struct ProxyConfig
 	int			keepidle;
 	int			keepintvl;
 	int			keepcnt;
+	char		default_user[NAMEDATALEN];
 } ProxyConfig;
 
-/* Single database connection */
-typedef struct
-{
-	struct ProxyCluster *cluster;
-	const char *connstr;		/* Connection string for libpq */
+typedef struct ConnUserInfo {
+	struct AANode node;
+	Oid user_oid;
+
+	char *username;
+	char *extra_connstr;		/* user= and password= */
+
+	SysCacheStamp umStamp;
+	bool needs_reload;
+} ConnUserInfo;
+
+typedef struct ProxyConnectionState {
+	struct AANode node;			/* node head in user->state tree */
+
+	ConnUserInfo *userinfo;
 
-	/* state */
 	PGconn	   *db;				/* libpq connection handle */
-	PGresult   *res;			/* last resultset */
-	int			pos;			/* Current position inside res */
 	ConnState	state;			/* Connection state */
 	time_t		connect_time;	/* When connection was started */
 	time_t		query_time;		/* When last query was sent */
 	bool		same_ver;		/* True if dest backend has same X.Y ver */
 	bool		tuning;			/* True if tuning query is running on conn */
+} ProxyConnectionState;
+
+/* Single database connection */
+typedef struct ProxyConnection
+{
+	struct AANode node;
+
+	struct ProxyCluster *cluster;
+	const char *connstr;		/* Connection string for libpq */
+
+	struct AATree userstate_tree; /* user->state tree */
+
+	/* state */
+	PGresult   *res;			/* last resultset */
+	int			pos;			/* Current position inside res */
+	ProxyConnectionState *cur;
 
 	/*
 	 * Nonzero if this connection should be used. The actual tag value is only
@@ -163,7 +207,7 @@ typedef struct
 /* Info about one cluster */
 typedef struct ProxyCluster
 {
-	struct ProxyCluster *next;	/* Pointer for building singly-linked list */
+	struct AANode node;			/* Node in name => cluster lookup tree */
 
 	const char *name;			/* Cluster name */
 	int			version;		/* Cluster version */
@@ -171,15 +215,23 @@ typedef struct ProxyCluster
 
 	int			part_count;		/* Number of partitions - power of 2 */
 	int			part_mask;		/* Mask to use to get part number from hash */
-	ProxyConnection **part_map; /* Pointers to conn_list */
+	ProxyConnection **part_map; /* Pointers to ProxyConnections */
 
-	int			conn_count;		/* Number of actual database connections */
-	ProxyConnection *conn_list; /* List of actual database connections */
+	int active_count;			/* number of active connections */
+	ProxyConnection **active_list; /* active ProxyConnection in current query */
+
+	struct AATree conn_tree;	/* connstr -> ProxyConnection */
+
+	struct AATree userinfo_tree; /* username->userinfo tree */
+	ConnUserInfo *cur_userinfo;	/* userinfo struct for current request */
 
 	int			ret_cur_conn;	/* Result walking: index of current conn */
 	int			ret_cur_pos;	/* Result walking: index of current row */
 	int			ret_total;		/* Result walking: total rows left */
 
+	Oid			sqlmed_server_oid;
+
+	bool		fake_cluster;	/* single connect-string cluster */
 	bool		sqlmed_cluster;	/* True if the cluster is defined using SQL/MED */
 	bool		needs_reload;	/* True if the cluster partition list should be reloaded */
 	bool		busy;			/* True if the cluster is already involved in execution */
@@ -188,8 +240,7 @@ typedef struct ProxyCluster
 	 * SQL/MED clusters: TIDs of the foreign server and user mapping catalog tuples.
 	 * Used in to perform cluster invalidation in syscache callbacks.
 	 */
-	ItemPointerData		clusterTupleId;
-	ItemPointerData		umTupleId;
+	SysCacheStamp clusterStamp;
 
 	/* notice processing: provide info about currently executing function */
 	struct ProxyFunction	*cur_func;
@@ -336,7 +387,8 @@ typedef struct ProxyFunction
 
 /* main.c */
 Datum		plproxy_call_handler(PG_FUNCTION_ARGS);
-void		plproxy_error(ProxyFunction *func, const char *fmt,...);
+void		plproxy_error(ProxyFunction *func, const char *fmt, ...)
+	__attribute__((format(PG_PRINTF_ATTRIBUTE, 2, 3)));
 void		plproxy_remote_error(ProxyFunction *func, ProxyConnection *conn, const PGresult *res, bool iserr);
 
 /* function.c */
@@ -351,6 +403,7 @@ ProxyFunction *plproxy_compile(FunctionCallInfo fcinfo, bool validate);
 /* execute.c */
 void		plproxy_exec(ProxyFunction *func, FunctionCallInfo fcinfo);
 void		plproxy_clean_results(ProxyCluster *cluster);
+void		plproxy_disconnect(ProxyConnectionState *cur);
 
 /* scanner.c */
 int			plproxy_yyget_lineno(void);
@@ -361,7 +414,8 @@ void		plproxy_yylex_startup(void);
 
 /* parser.y */
 void		plproxy_run_parser(ProxyFunction *func, const char *body, int len);
-void		plproxy_yyerror(const char *fmt,...);
+void		plproxy_yyerror(const char *fmt, ...)
+	__attribute__((format(PG_PRINTF_ATTRIBUTE, 1, 2)));
 
 /* type.c */
 ProxyComposite *plproxy_composite_info(ProxyFunction *func, TupleDesc tupdesc);
@@ -378,6 +432,7 @@ void		plproxy_cluster_cache_init(void);
 void		plproxy_syscache_callback_init(void);
 ProxyCluster *plproxy_find_cluster(ProxyFunction *func, FunctionCallInfo fcinfo);
 void		plproxy_cluster_maint(struct timeval * now);
+void		plproxy_activate_connection(struct ProxyConnection *conn);
 
 /* result.c */
 Datum		plproxy_result(ProxyFunction *func, FunctionCallInfo fcinfo);
